/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * res_config_pgsql.c <PostgreSQL plugin for portable configuration engine >
 * Copyright (C) 2005 Business Technology Group (http://www.btg.co.nz)
 *   Danel Swarbrick <daniel@btg.co.nz>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 0.20 $")

#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/config.h"
#include "asterisk/module.h"
#include "asterisk/lock.h"
#include "asterisk/options.h"
#include "asterisk/utils.h"

#include <libpq-fe.h>

static char *tdesc = "PostgreSQL Configuration";
STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

#define RES_CONFIG_PGSQL_CONF "res_pgsql.conf"
AST_MUTEX_DEFINE_STATIC(pgsql_lock);

char *pghost, *pgport, *dbName, *login, *pwd = NULL;
char *pgerror;
PGconn *conn;

int parse_config(void)
{
	struct ast_config *config;
	char *s;

	config = ast_config_load(RES_CONFIG_PGSQL_CONF);

	if (config) {

		// get the database host
		s = ast_variable_retrieve(config, "general", "dbhost");
		if (s == NULL) {
			ast_log(LOG_WARNING, "PgSQL RealTime: No database host found, using PostgreSQL default UNIX socket.\n");
			pghost = NULL;
		} else {
			pghost = strdup(s);
		}

		// get the database port
		s = ast_variable_retrieve(config, "general", "dbport");
		if (s == NULL) {
			ast_log(LOG_WARNING, "PgSQL RealTime: No database port found, assuming PostgreSQL default.\n");
			pgport = NULL;
		} else {
			pgport = strdup(s);
		}

		// get the database name
		s = ast_variable_retrieve(config, "general", "dbname");
		if (s == NULL) {
			ast_log(LOG_WARNING, "PgSQL RealTime: No database name found, assuming 'asterisk'.\n");
			dbName = "asterisk";
		} else {
			dbName = strdup(s);
		}

		// get the database user
		s = ast_variable_retrieve(config, "general", "dbuser");
		if (s == NULL) {
			ast_log(LOG_WARNING, "PgSQL RealTime: No database user found, assuming 'asterisk'.\n");
			login = "asterisk";
		} else {
			login = strdup(s);
		}

		// get the database password
		s = ast_variable_retrieve(config, "general", "dbpass");
		if (s == NULL) {
			ast_log(LOG_WARNING, "PgSQL RealTime: No database password found, assuming blank.\n");
			pwd = "";
		} else {
			pwd = strdup(s);
		}

	} else {
		ast_log(LOG_WARNING, "PgSQL RealTime config file not found\n");
	}
	ast_config_destroy(config);

	return 1;
}

PGconn *pgsqlReconnect(void)
{
	return conn;
}

static struct ast_variable *realtime_pgsql(const char *database, const char *table, va_list ap)
{
	PGresult *res;
	long int row, rowcount = 0;
	int col, colcount = 0;
	char sql[1024];
	char *stringp;
	char *chunk;
	char *op;	
	const char *newparam, *newval;
	struct ast_variable *var=NULL, *prev=NULL;

	if (!table) {
		ast_log(LOG_WARNING, "PgSQL RealTime: No table specified.\n");
		return NULL;
	}

	/* Get the first parameter and first value in our list of passed paramater/value pairs */
	newparam = va_arg(ap, const char *);
	newval = va_arg(ap, const char *);
	if (!newparam || !newval)  {
		ast_log(LOG_WARNING, "PgSQL RealTime: Realtime retrieval requires at least 1 parameter and 1 value to search on.\n");
		return NULL;
	}

	/* Create the first part of the query using the first parameter/value pairs we just extracted
	   If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat */

	if (!strchr(newparam, ' ')) op = " ="; else op = "";

	snprintf(sql, sizeof(sql), "SELECT * FROM %s WHERE %s%s '%s'", table, newparam, op, newval);
	while ((newparam = va_arg(ap, const char *))) {
		newval = va_arg(ap, const char *);
		if (!strchr(newparam, ' ')) op = " ="; else op = "";
		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " AND %s%s '%s'", newparam, op, newval);
	}
	va_end(ap);

	ast_log(LOG_NOTICE, "PgSQL RealTime: Retrieve SQL: %s\n", sql);

	ast_mutex_lock(&pgsql_lock);
	res = PQexec(conn, sql);

	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
		ast_log(LOG_WARNING, "PgSQL RealTime: Failed to query database. Check debug for more info.\n");
		ast_log(LOG_ERROR, "PgSQL RealTime: Query: %s\n", sql);
		ast_log(LOG_ERROR, "PgSQL RealTime: Query failed because: %s\n", PQresultErrorMessage(res));
		PQclear(res);
		ast_mutex_unlock(&pgsql_lock);
		return NULL;
	}

	rowcount = PQntuples(res);
	if (rowcount > 0) {
		colcount = PQnfields(res);

		for (row = 0; row < rowcount; row++) {
			for (col = 0; col < colcount; col++) {
				stringp = PQgetvalue(res, row, col);
				
				while (stringp) {
					chunk = strsep(&stringp, ";");
					if (chunk && !ast_strlen_zero(ast_strip(chunk))) {
						if (prev) {
							prev->next = ast_variable_new(PQfname(res, col), chunk);
							if (prev->next) {
								prev = prev->next;
							}
						} else {
							prev = var = ast_variable_new(PQfname(res, col), chunk);
						}
					}
				}
			}
		}
	} else {
		ast_log(LOG_WARNING, "PgSQL RealTime: Could not find any rows in table %s.\n", table);
	}

	PQclear(res);
	ast_mutex_unlock(&pgsql_lock);

	return var;
}

static struct ast_config *realtime_multi_pgsql(const char *database, const char *table, va_list ap)
{
	PGresult *res;
	long int row, rowcount = 0;
	int col, colcount = 0;
	char sql[1024];
	const char *initfield = NULL;
	char *stringp;
	char *chunk;
	char *op;
	const char *newparam, *newval;
	struct ast_realloca ra;
	struct ast_variable *var=NULL;
	struct ast_config *cfg = NULL;
	struct ast_category *cat = NULL;

	if(!table) {
		ast_log(LOG_WARNING, "PgSQL RealTime: No table specified.\n");
		return NULL;
	}
	
	memset(&ra, 0, sizeof(ra));

	cfg = ast_config_new();
	if (!cfg) {
		/* If I can't alloc memory at this point, why bother doing anything else? */
		ast_log(LOG_WARNING, "Out of memory!\n");
		return NULL;
	}

	/* Get the first parameter and first value in our list of passed paramater/value pairs */
	newparam = va_arg(ap, const char *);
	newval = va_arg(ap, const char *);
	if (!newparam || !newval)  {
		ast_log(LOG_WARNING, "PgSQL RealTime: Realtime retrieval requires at least 1 parameter and 1 value to search on.\n");
		return NULL;
	}

	initfield = ast_strdupa(newparam);
	if (initfield && (op = strchr(initfield, ' '))) {
		*op = '\0';
	}

	/* Create the first part of the query using the first parameter/value pairs we just extracted
	   If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat */

	if(!strchr(newparam, ' ')) op = " ="; else op = "";

	snprintf(sql, sizeof(sql), "SELECT * FROM %s WHERE %s%s '%s'", table, newparam, op, newval);
	while ((newparam = va_arg(ap, const char *))) {
		newval = va_arg(ap, const char *);
		if (!strchr(newparam, ' ')) op = " ="; else op = "";
		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " AND %s%s '%s'", newparam, op, newval);
	}

	if (initfield) {
		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " ORDER BY %s", initfield);
	}

	va_end(ap);

	ast_log(LOG_NOTICE, "PgSQL RealTime: Retrieve SQL: %s\n", sql);

	ast_mutex_lock(&pgsql_lock);
	res = PQexec(conn, sql);
	
	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
		ast_log(LOG_WARNING, "PgSQL RealTime: Failed to query database. Check debug for more info.\n");
		ast_log(LOG_NOTICE, "PgSQL RealTime: Query: %s\n", sql);
		ast_log(LOG_NOTICE, "PgSQL RealTime: Query failed because: %s\n", PQresultErrorMessage(res));
		PQclear(res);
		ast_mutex_unlock(&pgsql_lock);
		return NULL;
	}

	rowcount = PQntuples(res);
	if (rowcount > 0) {
		colcount = PQnfields(res);

		for (row = 0; row < rowcount; row++) {

			var = NULL;
			cat = ast_category_new("");
			if (!cat) {
				ast_log(LOG_WARNING, "Out of memory!\n");
				continue;
			}

			for (col = 0; col < colcount; col++) {
				stringp = PQgetvalue(res, row, col);
				while (stringp) {
					chunk = strsep(&stringp, ";");
					if (chunk && !ast_strlen_zero(ast_strip(chunk))) {
						if (initfield && !strcmp(initfield, PQfname(res, col))) {
							ast_category_rename(cat, chunk);
						}
						var = ast_variable_new(PQfname(res, col), chunk);
						ast_variable_append(cat, var);
					}
				}
			}
			ast_category_append(cfg, cat);
		}
	} else {
		ast_log(LOG_WARNING, "PgSQL RealTime: Could not find any rows in table %s.\n", table);
	}

	PQclear(res);
	ast_mutex_unlock(&pgsql_lock);

	return cfg;
}

static int update_pgsql(const char *database, const char *table, const char *keyfield, const char *lookup, va_list ap)
{
	PGresult *res;
	long int rowcount = 0;
	char sql[1024];
	const char *newparam, *newval;

	if (!table) {
		ast_log(LOG_WARNING, "PgSQL RealTime: No table specified.\n");
		return -1;
	}

	/* Get the first parameter and first value in our list of passed paramater/value pairs */
	newparam = va_arg(ap, const char *);
	newval = va_arg(ap, const char *);
	if (!newparam || !newval)  {
		ast_log(LOG_WARNING, "PgSQL RealTime: Realtime retrieval requires at least 1 parameter and 1 value to search on.\n");
		return -1;
	}

	/* Create the first part of the query using the first parameter/value pairs we just extracted
	   If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat */

	snprintf(sql, sizeof(sql), "UPDATE %s SET %s = '%s'", table, newparam, newval);
	while ((newparam = va_arg(ap, const char *))) {
		newval = va_arg(ap, const char *);
		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), ", %s = '%s'", newparam, newval);
	}
	va_end(ap);
	snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " WHERE %s = '%s'", keyfield, lookup);

	ast_log(LOG_NOTICE, "PgSQL RealTime: Update SQL: %s\n", sql);

	ast_mutex_lock(&pgsql_lock);
	res = PQexec(conn, sql);
	
	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
		ast_log(LOG_WARNING, "PgSQL RealTime: Failed to query database. Check debug for more info.\n");
		ast_log(LOG_NOTICE, "PgSQL RealTime: Query: %s\n", sql);
		ast_log(LOG_NOTICE, "PgSQL RealTime: Query failed because: %s\n", PQresultErrorMessage(res));
		PQclear(res);
		ast_mutex_unlock(&pgsql_lock);
		return -1;
	}

	rowcount = atol(PQcmdTuples(res));

	PQclear(res);
	ast_mutex_unlock(&pgsql_lock);

	ast_log(LOG_NOTICE, "PgSQL RealTime: Updated %lu rows on table: %s\n", rowcount, table);

	if (rowcount >= 0)
		return (int)rowcount;
	else
		return -1;
}

static struct ast_config *config_pgsql(const char *database, const char *table, const char *file, struct ast_config *cfg)
{
	PGresult *res;
	long int row, rowcount = 0;
	int colcount = 0;
	char sql[1024];
	char last[128] = "";
	struct ast_category *cur_cat;
	struct ast_variable *new_v;
	int last_cat_metric = 0;

	if (!file || !strcmp(file, RES_CONFIG_PGSQL_CONF)) {
		ast_log(LOG_WARNING, "PgSQL RealTime: Cannot configure myself.\n");
		return NULL;		
	}

	snprintf(sql, sizeof(sql), "SELECT category, var_name, var_val, cat_metric FROM %s WHERE filename='%s' and commented=0 ORDER BY filename, cat_metric desc, var_metric asc, category, var_name, var_val, id", table, file);

	ast_log(LOG_NOTICE, "PgSQL RealTime: Static SQL: %s\n", sql);

	ast_mutex_lock(&pgsql_lock);
	res = PQexec(conn, sql);
	
	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
		ast_log(LOG_WARNING, "PgSQL RealTime: Failed to query database. Check debug for more info.\n");
		ast_log(LOG_NOTICE, "PgSQL RealTime: Query: %s\n", sql);
		ast_log(LOG_NOTICE, "PgSQL RealTime: Query failed because: %s\n", PQresultErrorMessage(res));
		PQclear(res);
		ast_mutex_unlock(&pgsql_lock);
		return NULL;
	}

	rowcount = PQntuples(res);
	if (rowcount > 0) {

		ast_log(LOG_NOTICE, "PgSQL RealTime: Found %lu rows.\n", rowcount);
		colcount = PQnfields(res);
		for (row = 0; row < rowcount; row++) {
			if (!strcmp(PQgetvalue(res, row, 1), "#include")) {
				if (!ast_config_internal_load(PQgetvalue(res, row, 2), cfg)) {
					PQclear(res);
					ast_mutex_unlock(&pgsql_lock);
					return NULL;
				}
				continue;
			}

			if (strcmp(last, PQgetvalue(res, row, 0)) || last_cat_metric != atoi(PQgetvalue(res, row, 3))) {
				cur_cat = ast_category_new(PQgetvalue(res, row, 0));
				if (!cur_cat) {
					ast_log(LOG_WARNING, "Out of memory!\n");
					break;
				}
				strcpy(last, PQgetvalue(res, row, 0));
				last_cat_metric = atoi(PQgetvalue(res, row, 3));
				ast_category_append(cfg, cur_cat);
			}
			new_v = ast_variable_new(PQgetvalue(res, row, 1), PQgetvalue(res, row, 2));
			ast_variable_append(cur_cat, new_v);

		}

	} else {
		ast_log(LOG_WARNING, "PgSQL RealTime: Could not find config '%s' in database.\n", file);
	}

	PQclear(res);
	ast_mutex_unlock(&pgsql_lock);

	return cfg;
}

static struct ast_config_engine pgsql_engine = {
	.name = "pgsql",
	.load_func = config_pgsql,
	.realtime_func = realtime_pgsql,
	.realtime_multi_func = realtime_multi_pgsql,
	.update_func = update_pgsql
};

int unload_module(void)
{
	// acquire control before doing anything to the module itself
	ast_mutex_lock(&pgsql_lock);

	PQfinish(conn);
	ast_config_engine_deregister(&pgsql_engine);
	if (option_verbose) {
		ast_verbose("PgSQL RealTime unloaded.\n");
	}

	STANDARD_HANGUP_LOCALUSERS;

	// Unlock so something else can destroy the lock.
	ast_mutex_unlock(&pgsql_lock);

	return 0;
}

int load_module(void)
{
	parse_config();

	conn = PQsetdbLogin(pghost, pgport, NULL, NULL, dbName, login, pwd);

	if (PQstatus(conn) != CONNECTION_BAD) {
		ast_log(LOG_NOTICE, "PgSQL RealTime: Successfully connected to PostgreSQL database.\n");
	} else {
		pgerror = PQerrorMessage(conn);
		ast_log(LOG_WARNING, "PgSQL RealTime: Couldn't establish DB connection. Check debug.\n");
		ast_log(LOG_ERROR, "PgSQL RealTime: reason %s\n", pgerror);
	}

	ast_config_engine_register(&pgsql_engine);
	if (option_verbose) {
		ast_verbose("PgSQL RealTime driver loaded.\n");
	}

	return 0;
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	/* never unload a config module */
	return 1;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
